"""
Module for classes related to the solution of inverse problems via optimisation methods. 
"""

from scipy.stats import chi2
from pygeoinf.hilbert_space import Operator
from pygeoinf.inversion import Inversion
from pygeoinf.linear_solvers import IterativeLinearSolver


class LinearLeastSquaresInversion(Inversion):
    """
    Class for the solution of regularised linear least-squares problems within a Hilbert space.
    """

    def __init__(self, forward_problem):
        """
        Args:
            forward_problem (LinearForwardProblem): The forward problem. If the problem has
                a data errror measure, then this measure must have its inverse covariance
                operator set.
        """
        super().__init__(forward_problem)
        self.assert_inverse_data_covariance()

    def normal_operator(self, damping):
        """
        Returns the least-squares normal operator.

        Args:
            damping (float): The norm damping parameter. Must be non-negative.

        Returns:
            LinearOperator: The normal operator.

        Raises:
            ValueError: If damping is not non-negative.
        """
        if damping < 0:
            raise ValueError("Damping parameter must be non-negative.")
        forward_operator = self.forward_problem.forward_operator
        identity = self.forward_problem.model_space.identity_operator()

        if self.forward_problem.data_error_measure_set:

            inverse_data_covariance = (
                self.forward_problem.data_error_measure.inverse_covariance
            )
            return (
                forward_operator.adjoint @ inverse_data_covariance @ forward_operator
                + damping * identity
            )
        else:
            return forward_operator.adjoint @ forward_operator + damping * identity

    def least_squares_operator(self, damping, solver, /, *, preconditioner=None):
        """
        Returns an operator that maps data to the least squares solution.

        Args:
            damping (float): The norm damping parameter. Must be non-negative.
            solver (LinearSolver): Linear solver for the normal equations.
            preconditioner (LinearOperator): Preconditioner for use in
                solving the normal equations. The default is None.

        Returns:
            Operator: Mapping from data space to least-squares solution.

        Raises:
            ValueError: If damping is not non-negative.
            ValueError: If solver is not a instance of LinearSolver.

        Notes. If the data is error-free the object returned is a LinearOperator.
        """

        forward_operator = self.forward_problem.forward_operator
        normal_operator = self.normal_operator(damping)

        if isinstance(solver, IterativeLinearSolver):
            inverse_normal_operator = solver(
                normal_operator, preconditioner=preconditioner
            )
        else:
            inverse_normal_operator = solver(normal_operator)

        if self.forward_problem.data_error_measure_set:
            inverse_data_covariance = (
                self.forward_problem.data_error_measure.inverse_covariance
            )

            def mapping(data):
                shifted_data = self.forward_problem.data_space.subtract(
                    data, self.forward_problem.data_error_measure.expectation
                )
                return (
                    inverse_normal_operator
                    @ forward_operator.adjoint
                    @ inverse_data_covariance
                )(shifted_data)

            return Operator(self.data_space, self.model_space, mapping)

        else:
            return inverse_normal_operator @ forward_operator.adjoint

    def model_measure(
        self,
        damping,
        data,
        solver,
        /,
        *,
        preconditioner=None,
    ):
        """
        Returns the measure on the model space induced by the observed data under the least-squares solution.

        Args:
            damping (float): The norm damping parameter. Must be non-negative.
            data (data vector): Observed data
            solver (LinearSolver): Linear solver for solvint the normal equations.
            preconditioner (LinearOperator): Preconditioner for use in
                solving the normal equations. The default is the identity.

        Returns:
            GaussianMeasure: Measure on the model space induced by the
                least-squares solution for given data. Note that this measure only
                accounts for uncertainty due to the propagation of
                uncertainties within the data.

        Raises:
            ValueError: If damping is not non-negative.
            ValueError: If solver is not a instance of LinearSolver.
        """

        if not self.forward_problem.data_error_measure_set:
            raise NotImplementedError("Data error measure not set")

        least_squares_operator = self.least_squares_operator(
            damping,
            solver,
            preconditioner=preconditioner,
        )
        model = least_squares_operator(data)
        return self.forward_problem.data_error_measure.affine_mapping(
            operator=least_squares_operator, translation=model
        )


class LinearMinimumNormInversion(Inversion):
    """
    Class for the solution of linear minimum norm problems in a Hilbert space.
    """

    def __init__(self, forward_problem):
        """
        Args:
            forward_problem (LinearForwardProblem): The forward problem. If the problem has
                a data errror measure, then this measure must have its inverse covariance
                operator set.
        """
        super().__init__(forward_problem)
        self.assert_inverse_data_covariance()

    def minimum_norm_operator(
        self,
        solver,
        /,
        *,
        preconditioner=None,
        significance_level=0.05,
        minimum_damping=0,
        maxiter=100,
        rtol=1.0e-6,
        atol=0,
    ):
        """
        Return the operator that maps data to the minimum norm solution.
        In the case of error-free data this operator is linear, but once
        data errors are present it is a non-linear opearator.

        Args:
            solver (LinearSolver): Solver for solution of the necessary linear systems.
            preconditioner (LinearOperator): Preconditioner for use with iterative solvers.
            significance_level (float): Significance level used to set the crtical value
                for chi-squared. Must lie in the interval (0,1). Default is 0.05 which
                corresponds to 95% acceptance.
            minimum_damping (float): Minimum value of the damping when looking for a
                lower bound on the damping. Default is 0.
            maxiter (int): Maximum number of iterations within the bracketing method.
                Default is 100.
            rtol (float): Relative tolerance within the bracketing step. Default is 1e-6
            atyol (float): Absolute tolerance within the bracketing step. Default is 0.
        """

        if self.forward_problem.data_error_measure_set:

            critical_value = self.forward_problem.crtical_chi_squared(
                significance_level
            )
            least_squares_inversion = LinearLeastSquaresInversion(self.forward_problem)

            def least_squares_mapping(damping, data, model0=None):

                forward_operator = self.forward_problem.forward_operator
                normal_operator = least_squares_inversion.normal_operator(damping)
                inverse_data_covariance = (
                    self.forward_problem.data_error_measure.inverse_covariance
                )

                shifted_data = self.data_space.subtract(
                    data, self.forward_problem.data_error_measure.expectation
                )

                if isinstance(solver, IterativeLinearSolver):
                    model_in = (forward_operator.adjoint @ inverse_data_covariance)(
                        shifted_data
                    )
                    model = solver.solve_linear_system(
                        normal_operator,
                        preconditioner,
                        model_in,
                        x0=model0,
                    )

                else:
                    inverse_normal_operator = solver(normal_operator)
                    model = (
                        inverse_normal_operator
                        @ forward_operator.adjoint
                        @ inverse_data_covariance
                    )(shifted_data)

                chi_squared = self.forward_problem.chi_squared(model, data)
                return chi_squared, model

            def mapping(data):

                model = self.model_space.zero
                chi_squared = self.forward_problem.chi_squared(model, data)
                if chi_squared <= critical_value:
                    return model

                damping = 1
                chi_squared, _ = least_squares_mapping(damping, data)

                damping_lower = damping if chi_squared <= critical_value else None
                damping_upper = damping if chi_squared > critical_value else None

                if damping_lower is None:
                    it = 0
                    while chi_squared > critical_value:
                        it += 1
                        damping /= 2
                        chi_squared, _ = least_squares_mapping(damping, data)
                        if chi_squared < minimum_damping or it == maxiter:
                            raise RuntimeError("Crtical value cannot be obtained")
                    damping_lower = damping

                if damping_upper is None:
                    while chi_squared < critical_value:
                        damping *= 2
                        chi_squared, _ = least_squares_mapping(damping, data)
                    damping_upper = damping

                model0 = None
                for _ in range(maxiter):

                    damping = 0.5 * (damping_lower + damping_upper)
                    chi_squared, model = least_squares_mapping(damping, data, model0)
                    model0 = self.model_space.copy(model)

                    if chi_squared < critical_value:
                        damping_lower = damping
                    else:
                        damping_upper = damping

                    if damping_upper - damping_lower < atol + rtol * (
                        damping_lower + damping_upper
                    ):
                        return model

                raise RuntimeError("Bracketing has failed")

            return Operator(self.model_space, self.data_space, mapping)

        else:
            forward_operator = self.forward_problem.forward_operator
            normal_operator = forward_operator @ forward_operator.adjoint
            inverse_normal_operator = solver(normal_operator)
            return forward_operator.adjoint @ inverse_normal_operator



"""
Sobolev spaces for functions on a line. 
"""

import matplotlib.pyplot as plt
import numpy as np
from pygeoinf.hilbert_space import (
    HilbertSpace,
    LinearOperator,
    EuclideanSpace,
)
from pygeoinf.gaussian_measure import GaussianMeasure
from pygeoinf.homogeneous_space.circle import Sobolev as CicleSobolev


class Sobolev(HilbertSpace):
    """
    Implementation of the Sobolev space H^s on a line.
    """

    def __init__(
        self,
        kmax,
        exponent,
        length_scale,
        /,
        *,
        x0=0,
        x1=1,
    ):
        """
        Args:
            kmax (float): The maximum Fourier degree.
            exponent (float): Sobolev exponent.
            length_scale (float): Sobolev length-scale.
            radius (float): Radius of the circle. Default is 1.
            x0 (float): The left boudary of the interval, Default is 0.
            x1 (float): The right boundary of the interval. Default is 1.
        """

        if x0 >= x1:
            raise ValueError("Invalid interval parameters")

        if exponent != 0 and length_scale <= 0:
            raise ValueError("Length-scale must be positive")

        self._kmax = kmax
        self._x0 = x0
        self._x1 = x1
        self._exponent = exponent
        self._length_scale = length_scale

        # Work out transformation from a unit circle.
        padding_scale = 5 * length_scale if length_scale > 0 else 0.1 * (x1 - x0)
        number_of_points = 2 * kmax
        width = x1 - x0
        self._start_index = int(
            number_of_points * padding_scale / (width + 2 * padding_scale)
        )
        self._finish_index = 2 * kmax - self._start_index + 1
        self._padding_length = (
            self._start_index * width / (number_of_points - 2 * self._start_index)
        )

        self._jac = (width + 2 * self._padding_length) / (2 * np.pi)
        self._ijac = 1 / self._jac
        self._sqrt_jac = np.sqrt(self._jac)
        self._isqrt_jac = 1 / self._sqrt_jac

        # Set up the related Sobolev space on the unit circle.
        circle_length_scale = length_scale * self._ijac
        self._circle_space = CicleSobolev(kmax, exponent, circle_length_scale)

        super().__init__(
            self._circle_space.dim,
            self._to_components,
            self._from_components,
            self._inner_product,
            self._to_dual,
            self._from_dual,
            vector_multiply=lambda u1, u2: u1 * u2,
        )

    @property
    def kmax(self):
        """
        Return the maximum Fourier degree.
        """
        return self._kmax

    @property
    def x0(self):
        """
        Returns the left boundary point.
        """
        return self._x0

    @property
    def x1(self):
        """
        Returns the right boundary point.
        """
        return self._x1

    @property
    def width(self):
        """
        Return the radius.
        """
        return self._x1 - self._x0

    @property
    def exponent(self):
        """
        Return the Sobolev exponent.
        """
        return self._exponent

    @property
    def length_scale(self):
        """
        Return the relative Sobolev length_scale.
        """
        return self._length_scale

    @property
    def point_spacing(self):
        """
        Return the point spacing.
        """
        return self._circle_space.angle_spacing * self._jac

    def computational_points(self):
        """
        Returns a numpy array of the computational points.
        """
        return self._x0 - self._padding_length + self._jac * self._circle_space.angles()

    def points(self):
        """
        Returns a numpy array of the points.
        """
        return self.computational_points()[self._start_index : self._finish_index]

    def project_function(self, f):
        """
        Returns an element of the space formed by projecting a given function.
        """
        return np.fromiter(
            [f(x) * self._taper(x) for x in self.computational_points()], float
        )

    def plot(self, u, fig=None, ax=None, computational_domain=False, **kwargs):
        """
        Make a simple plot of an element of the space on the computational domain.

        Args:
            u (vector): The element of the space to be plotted.
            fig (Figure): An existing Figure object to use. Default is None.
            ax (Axes): An existing Axes object to use. Default is None.
            computatoinal_domain (bool): If True, plot the whole computational
                domain. Default is False.
            kwargs: Keyword arguments forwarded to plot.

        Returns
            Figure: The figure object, either that given or newly created.
            Axes: The axes object, either that given or newly created.
        """

        figsize = kwargs.pop("figsize", (10, 8))

        if fig is None:
            fig = plt.figure(figsize=figsize)
        if ax is None:
            ax = fig.add_subplot()

        if computational_domain:
            ax.plot(self.computational_points(), u, **kwargs)
        else:
            ax.plot(self.points(), u[self._start_index : self._finish_index], **kwargs)

        return fig, ax

    def plot_pointwise_bounds(
        self, u, u_bound, fig=None, ax=None, computational_domain=False, **kwargs
    ):
        """
        Make a plot of an element of the space bounded above and below by a standard
        deviation curve.

        Args:
            u (vector): The element of the space to be plotted.
            u_bounds (vector): A second element giving point-wise bounds.
            fig (Figure): An existing Figure object to use. Default is None.
            ax (Axes): An existing Axes object to use. Default is None.
            kwargs: Keyword arguments forwarded to plot.

        Returns
            Figure: The figure object, either that given or newly created.
            Axes: The axes object, either that given or newly created.
        """

        figsize = kwargs.pop("figsize", (10, 8))

        if fig is None:
            fig = plt.figure(figsize=figsize)
        if ax is None:
            ax = fig.add_subplot()

        if computational_domain:
            ax.fill_between(
                self.computational_points(), u - u_bound, u + u_bound, **kwargs
            )
        else:
            ax.fill_between(
                self.points(),
                u[self._start_index : self._finish_index]
                - u_bound[self._start_index : self._finish_index],
                u[self._start_index : self._finish_index]
                + u_bound[self._start_index : self._finish_index],
                **kwargs,
            )

        return fig, ax

    def invariant_automorphism(self, f):
        """
        Returns a linear operator on the space of the form f(Delta) with Delta the Laplacian.
        The resulting operator is only well-defined as a continuous automorphism if f is bounded.
        """
        A = self._circle_space.invariant_automorphism(lambda k: self._ijac * k)
        return LinearOperator.formally_self_adjoint(self, A)

    def invariant_gaussian_measure(self, f, /, *, expectation=None):
        """
        Returns a Gaussian measure on the space whose covariance takes the form f(Delta) with
        Delta the Laplacian. The trace-class condition for the measure to be well-defined implies
        that the sequence {f(k)} must be summable.
        """
        mu = self._circle_space.invariant_gaussian_measure(
            lambda k: self._ijac * k, expectation=expectation
        )
        covariance = LinearOperator.self_adjoint(self, mu.covariance)
        return GaussianMeasure(
            covariance=covariance, expectation=expectation, sample=mu.sample
        )

    def sobolev_gaussian_measure(
        self, exponent, length_scale, amplitude, /, *, expectation=None
    ):
        """
        Returns a Gaussian measure with Sobolev covariance. The measure is
        scaled such that the its pointwise standard deviation is equal
        to the optional amplitude (default value 1).
        """
        mu = self._circle_space.sobolev_gaussian_measure(
            exponent, length_scale * self._ijac, amplitude, expectation=expectation
        )
        covariance = LinearOperator.self_adjoint(self, mu.covariance)
        return GaussianMeasure(
            covariance=covariance, expectation=expectation, sample=mu.sample
        )

    def heat_gaussian_measure(self, length_scale, amplitude, /, *, expectation=None):
        """
        Returns a Gaussian measure with heat kernel covariance. The measure is
        scaled such that the its pointwise standard deviation is equal
        to the optional amplitude (default value 1).
        """
        mu = self._circle_space.heat_gaussian_measure(
            length_scale * self._ijac, amplitude, expectation=expectation
        )
        covariance = LinearOperator.self_adjoint(self, mu.covariance)
        return GaussianMeasure(
            covariance=covariance, expectation=expectation, sample=mu.sample
        )

    def dirac(self, x):
        """
        Returns a dirac measure as a LinearForm on the space.
        """
        theta = self._inverse_transformation(x)
        up = self._circle_space.dirac(theta)
        cp = self._circle_space.dual.to_components(up) * self._isqrt_jac
        return self.dual.from_components(cp)

    def dirac_representation(self, x):
        """
        Returns the representation of the dirac measure based at the point.
        """
        return self.from_dual(self.dirac(x))

    def point_evaluation_operator(self, points):
        """
        Returns the point evaluation operator on the space
        at the given list, xs, of points.
        """
        dim = len(points)
        matrix = np.zeros((dim, self.dim))

        for i, point in enumerate(points):
            cp = self.dirac(point).components
            matrix[i, :] = cp

        return LinearOperator.from_matrix(
            self, EuclideanSpace(dim), matrix, galerkin=True
        )

    # =============================================================#
    #                        Private methods                       #
    # =============================================================#

    def _step(self, x):
        if x > 0:
            return np.exp(-1 / x)
        else:
            return 0

    def _bump_up(self, x, x1, x2):
        s1 = self._step(x - x1)
        s2 = self._step(x2 - x)
        return s1 / (s1 + s2)

    def _bump_down(self, x, x1, x2):
        s1 = self._step(x2 - x)
        s2 = self._step(x - x1)
        return s1 / (s1 + s2)

    def _taper(self, x):
        s1 = self._bump_up(x, self._x0 - self._padding_length, self._x0)
        s2 = self._bump_down(x, self._x1, self._x1 + self._padding_length)
        return s1 * s2

    def _transformation(self, th):
        return self._x0 - self._padding_length + self._jac * th

    def _inverse_transformation(self, x):
        return (x - self._x0 + self._padding_length) * self._ijac

    def _to_components(self, u):
        c = self._circle_space.to_components(u)
        c *= self._sqrt_jac
        return c

    def _from_components(self, c):
        c *= self._isqrt_jac
        u = self._circle_space.from_components(c)
        return u

    def _inner_product(self, u1, u2):
        return self._jac * self._circle_space.inner_product(u1, u2)

    def _to_dual(self, u):
        up = self._circle_space.to_dual(u)
        cp = self._circle_space.dual.to_components(up) * self._sqrt_jac
        return self.dual.from_components(cp)

    def _from_dual(self, up):
        cp = self.dual.to_components(up)
        vp = self._circle_space.dual.from_components(cp) * self._isqrt_jac
        return self._circle_space.from_dual(vp)



"""
Module for Gaussian measures on Hilbert spaces. 
"""

import numpy as np
from scipy.linalg import eigh
from scipy.sparse import diags
from scipy.stats import multivariate_normal
from pygeoinf.hilbert_space import (
    LinearOperator,
    DiagonalLinearOperator,
    EuclideanSpace,
)
from pygeoinf.direct_sum import (
    BlockDiagonalLinearOperator,
)


class GaussianMeasure:
    """
    Class for Gaussian measures on a Hilbert space.
    """

    def __init__(
        self,
        /,
        *,
        covariance=None,
        covariance_factor=None,
        expectation=None,
        sample=None,
        inverse_covariance=None,
        inverse_covariance_factor=None,
    ):
        """
        Args:
            domain (HilbertSpace): The Hilbert space on which the measure is defined.
            covariance (LinearOperator): A self-adjoint and non-negative linear operator on the domain.
                Default is None.
            covariance_factor (LinearOperator): A linear operator from Euclidean space to the domain of
                the measure such that C = LL*. Default is None.
            expectation (vector): The expectation value of the measure. Default is zero.
            sample (callable): A functor that returns a random sample from the measure. Default is None.
            inverse_covariance (LinearOperator | None): The inverse of the covariance. Default is None.
            inverse_covariance_factor (LinearOperator | None): The inverse of the covariance. Default is None.

        Notes:
            Either the covariance or covariance factor must be specified, but not both. In the covariance
            factor is provided, then a sampling method is generated by default. Otherwise one can optionally
            be provided.
        """

        if covariance is None and covariance_factor is None:
            raise ValueError(
                "Neither covariance or covariance factor has been provided"
            )

        self._covariance_factor = covariance_factor
        self._covariance = (
            covariance_factor @ covariance_factor.adjoint
            if covariance is None
            else covariance
        )
        self._domain = self._covariance.domain
        self._sample = sample if covariance_factor is None else self._sample_from_factor
        self._inverse_covariance_factor = inverse_covariance_factor
        if inverse_covariance_factor is not None:
            self._inverse_covariance = (
                inverse_covariance_factor.adjoint @ inverse_covariance_factor
            )
        elif inverse_covariance is not None:
            self._inverse_covariance = inverse_covariance
        else:
            self._inverse_covariance = None

        if expectation is None:
            self._expectation = self.domain.zero
        else:
            self._expectation = expectation

    @staticmethod
    def from_standard_deviation(domain, standard_deviation, /, *, expectation=None):
        """
        Forms a Gaussian measure on a Hilbert space with zero
        expectation and whose covariance is proportional to the
        identity operator.

        Args:
            domain (HilbertSpace): The Hilbert space on which the measure
                is defined.
            standard_devitation (float): The standard deviation by which the
                identity is scaled to form the covariance.
            expectation (vector): The expectation of the measure, Default is
                None which corresponds to the zero-vector.

        Notes:
            This measure is only well-defined for finite-dimensional spaces and not
            for finite-dimensional approximations to infinite-dimensional spaces.
        """
        covariance_factor = standard_deviation * domain.identity_operator()
        inverse_covariance_factor = (
            1 / standard_deviation
        ) * domain.identity_operator()
        return GaussianMeasure(
            covariance_factor=covariance_factor,
            inverse_covariance_factor=inverse_covariance_factor,
            expectation=expectation,
        )

    @staticmethod
    def from_standard_deviations(domain, standard_deviations, /, *, expectation=None):
        """
        Forms a Gaussian measure on a Hilbert space with zero
        expectation and whose covariance is diagonal within the
        Galerkin representation.

        Args: domain (HilbertSpace): The Hilbert space on which the measure
                is defined.
        standard_devitations (numpy vector): The diagonal values of the covariance
            within its Galerkin representation.
        expectation (vector): The expectation of the measure, Default is
                None which corresponds to the zero-vector.

        Raises:
            ValueError: If the dimension of the standard deviation vector does not
                match that of the Hilbert space.

        """
        if standard_deviations.size != domain.dim:
            raise ValueError(
                "Standard deviation vector does not have the correct length"
            )
        euclidean = EuclideanSpace(domain.dim)
        covariance_factor = DiagonalLinearOperator(
            euclidean, domain, standard_deviations, galerkin=True
        )
        return GaussianMeasure(
            covariance_factor=covariance_factor,
            inverse_covariance_factor=covariance_factor.inverse,
            expectation=expectation,
        )

    @staticmethod
    def from_covariance_matrix(domain, covariance_matrix, /, *, expectation=None):
        """
        Forms a Gaussian measure from the Galerking representation
        of the covariance matrix.
        """

        eigenvalues, U = eigh(covariance_matrix)
        if any(val < 0 for val in eigenvalues):
            raise ValueError("Covariance matrix is not non-negative")

        values = np.fromiter([np.sqrt(val) for val in eigenvalues], dtype=float)
        D = diags([values], [0])
        Di = diags([np.reciprocal(values)], [0])
        L = U @ D
        Li = Di @ U.T

        covariance_factor = LinearOperator.from_matrix(
            EuclideanSpace(domain.dim), domain, L
        )
        inverse_covariance_factor = LinearOperator.from_matrix(
            domain, EuclideanSpace(domain.dim), Li
        )

        return GaussianMeasure(
            covariance_factor=covariance_factor,
            inverse_covariance_factor=inverse_covariance_factor,
            expectation=expectation,
        )

    @staticmethod
    def from_samples(domain, samples):
        """
        Forms a Gaussian measure from a set of samples based on
        the sample expectation and sample variance.

        Args:
            domain (HilbertSpace): The space the measure is defined on.
            samples ([vectors]): A list of samples.

        Notes:
            :
        """
        assert all([domain.is_element(x) for x in samples])

        n = len(samples)
        expectation = domain.zero
        for x in samples:
            expectation = domain.axpy(1 / n, x, expectation)

        if n == 1:
            covariance = LinearOperator.self_adjoint(domain, lambda x: domain.zero)

            def sample():
                return expectation

        else:
            offsets = [domain.subtract(x, expectation) for x in samples]
            covariance = LinearOperator.self_adjoint_from_tensor_product(
                domain, offsets
            ) / (n - 1)

            def sample():
                x = expectation
                randoms = np.random.randn(len(offsets))
                for y, r in zip(offsets, randoms):
                    x = domain.axpy(r / np.sqrt(n - 1), y, x)
                return x

        return GaussianMeasure(
            covariance=covariance, expectation=expectation, sample=sample
        )

    @staticmethod
    def from_direct_sum(measures):
        """
        Forms the direct sum of a list of Gaussian measures.
        """

        expectation = [measure.expectation for measure in measures]

        covariance = BlockDiagonalLinearOperator(
            [measure.covariance for measure in measures]
        )

        inverse_covariance = (
            BlockDiagonalLinearOperator(
                [measure.inverse_covariance for measure in measures]
            )
            if all(measure.inverse_covariance_set for measure in measures)
            else None
        )

        def sample_impl():
            return [measure.sample() for measure in measures]

        sample = (
            sample_impl if all(measure.sample_set for measure in measures) else None
        )

        return GaussianMeasure(
            covariance=covariance,
            expectation=expectation,
            sample=sample,
            inverse_covariance=inverse_covariance,
        )

    @property
    def domain(self):
        """The Hilbert space the measure is defined on."""
        return self._domain

    @property
    def covariance(self):
        """The covariance operator as an instance of LinearOperator."""
        return self._covariance

    @property
    def inverse_covariance_set(self):
        """Returns true if inverse covariance is available"""
        return self._inverse_covariance is not None

    @property
    def inverse_covariance(self):
        """The covariance operator as an instance of LinearOperator."""
        if self._inverse_covariance is None:
            raise NotImplementedError("Inverse covariance is not set")
        else:
            return self._inverse_covariance

    @property
    def covariance_factor_set(self):
        """
        True if covariance factor set.
        """
        return self._covariance_factor is not None

    @property
    def covariance_factor(self):
        """
        Return the covariance factor if set.
        """
        if not self.covariance_factor_set:
            raise NotImplementedError("Covariance factor has not been set")
        return self._covariance_factor

    @property
    def inverse_covariance_factor_set(self):
        """
        True if inverse covariance factor set.
        """
        return self._inverse_covariance_factor is not None

    @property
    def inverse_covariance_factor(self):
        """
        Return the inverse covariance factor if set.
        """
        if not self.covariance_factor_set:
            raise NotImplementedError("Covariance factor has not been set")
        return self._inverse_covariance_factor

    @property
    def expectation(self):
        """The expectation of the measure."""
        return self._expectation

    @property
    def sample_set(self):
        """
        Returns true if sample method is available.
        """
        return self._sample is not None

    def sample(self):
        """
        Returns a random sample drawn from the measure.
        """
        if self._sample is None:
            raise NotImplementedError("Sample method is not set.")
        else:
            return self._sample()

    def samples(self, n):
        """
        Returns a list of n samples form the measure.
        """
        assert n >= 1
        return [self.sample() for _ in range(n)]

    def sample_expectation(self, n):
        """
        Returns the sample expectation using n > 1 samples.
        """
        assert n >= 1
        expectation = self.domain.zero
        for _ in range(n):
            sample = self.sample()
            expectation = self.domain.axpy(1 / n, sample, expectation)
        return expectation

    def sample_pointwise_variance(self, n):
        """
        Returns the sample variance using n > 1 samples.
        Available only if multiplication of vectors is defined
        on the domain of the measure.
        """
        if not self.domain.has_vector_multiply:
            raise NotImplementedError("Method not defined for this measure")

        samples = self.samples(n)
        expectation = self.expectation
        variance = self.domain.zero

        for sample in samples:
            diff = self.domain.subtract(sample, expectation)
            prod = self.domain.vector_multiply(diff, diff)
            variance = self.domain.axpy(1 / n, prod, variance)

        return variance

    def affine_mapping(self, /, *, operator=None, translation=None):
        """
        Returns the push forward of the measure under an affine mapping.

        Args:
            operator (LinearOperator): The operator part of the mapping.
            translation (vector): The translational part of the mapping.

        Returns:
            GaussianMeasure: The transformed measure defined on the
                codomain of the operator.

        Raises:
            ValueError: If the domain of the operator domain is not
                the domain of the measure.

        Notes:
            If operator is not set, it defaults to the identity.
            It translation is not set, it defaults to zero.
        """

        if operator is None:
            _operator = self.domain.identity_operator()
        else:
            _operator = operator

        if translation is None:
            _translation = _operator.codomain.zero
        else:
            _translation = translation

        expectation = _operator.codomain.add(_operator(self.expectation), _translation)

        if self.covariance_factor_set:

            covariance_factor = _operator @ self.covariance_factor

            return GaussianMeasure(
                covariance_factor=covariance_factor, expectation=expectation
            )

        else:

            covariance = _operator @ self.covariance @ _operator.adjoint

            def sample():
                return _operator.codomain.add(_operator(self.sample()), _translation)

            return GaussianMeasure(
                covariance=covariance, expectation=expectation, sample=sample
            )

    def as_multivariate_normal(self):
        """
        Returns the measure in the form of a scipy.stats multivariate_normal
        distribution. This method is available only if the measure is
        defined on a EuclideanSpace.
        """

        if not isinstance(self.domain, EuclideanSpace):
            raise NotImplementedError(
                "Method only defined for measures on Euclidean space"
            )

        return multivariate_normal(
            mean=self.expectation, cov=self.covariance.matrix(dense=True)
        )

    def low_rank_approximation(
        self,
        rank,
        /,
        *,
        power=0,
        method="fixed",
        rtol=1e-2,
        scaling=None,
    ):
        """
        Returns an approximation to the measure with the covariance operator
        replaced by a low-rank Cholesky decomposition along with the associated
        sampling method.
        """

        covariance = self.covariance

        if scaling is None:
            covariance_factor = covariance.random_cholesky(
                rank, power=power, method=method, rtol=rtol
            )

        else:
            covariance = scaling @ covariance @ scaling
            covariance_factor = covariance.random_cholesky(
                rank, power=power, method=method, rtol=rtol
            )
            covariance_factor = scaling.inverse @ covariance_factor

        return GaussianMeasure(
            covariance_factor=covariance_factor,
            expectation=self.expectation,
        )

    def _sample_from_factor(self):
        covariance_factor = self.covariance_factor
        value = covariance_factor(np.random.randn(self.covariance_factor.domain.dim))
        return (
            self.domain.add(value, self.expectation)
            if self.expectation is not None
            else value
        )

    def __neg__(self):
        """Negative of the measure."""

        if self.covariance_factor_set:
            return GaussianMeasure(
                covariance_factor=self.covariance_factor,
                expectation=self.domain.negative(self.expectation),
            )
        else:
            return GaussianMeasure(
                covariance=self.covariance,
                expectation=self.domain.negative(self.expectation),
                sample=lambda: self.domain.negative(self.sample()),
            )

    def __mul__(self, alpha):
        """Multiply the measure by a scalar."""
        if self.covariance_factor_set:
            return GaussianMeasure(
                covariance_factor=alpha * self.covariance_factor,
                expectation=self.domain.multiply(alpha, self.expectation),
            )
        return GaussianMeasure(
            covariance=alpha * alpha * self.covariance,
            expectation=self.domain.multiply(alpha, self.expectation),
            sample=lambda: self.domain.multiply(alpha, self.sample()),
        )

    def __rmul__(self, alpha):
        """Multiply the measure by a scalar."""
        return self * alpha

    def __add__(self, other):
        """Add two measures on the same domain."""
        return GaussianMeasure(
            covariance=self.covariance + other.covariance,
            expectation=self.domain.add(self.expectation, other.expectation),
            sample=lambda: self.domain.add(self.sample(), other.sample()),
        )

    def __sub__(self, other):
        """Subtract two measures on the same domain."""
        return GaussianMeasure(
            covariance=self.covariance + other.covariance,
            expectation=self.domain.subtract(self.expectation, other.expectation),
            sample=lambda: self.domain.subtract(self.sample(), other.sample()),
        )



"""
Module for Sobolev spaces on the two-sphere.
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.sparse import diags, coo_array
import pyshtools as sh


from pygeoinf.hilbert_space import (
    HilbertSpace,
    LinearOperator,
    EuclideanSpace,
    LinearForm,
)

from pygeoinf.gaussian_measure import GaussianMeasure


class SHToolsHelper:
    """Helper class for working with pyshtool grid functions and coefficients."""

    def __init__(self, lmax, /, *, radius=1, grid="DH"):
        """
        Args:
            lmax (int): Maximum degree for spherical harmonic expansions.
            radius (float): radius of the sphere.
            grid (str): Grid type for spatial functions.
        """
        self._lmax = lmax
        self._radius = radius
        self._grid = grid
        if self.grid == "DH2":
            self._sampling = 2
        else:
            self._sampling = 1
        self._extend = True
        self._normalization = "ortho"
        self._csphase = 1
        self._sparse_coeffs_to_component = self._coefficient_to_component_mapping()

    @property
    def lmax(self):
        """Returns truncation degree."""
        return self._lmax

    @property
    def dim(self):
        """Returns dimension of vector space."""
        return (self.lmax + 1) ** 2

    @property
    def radius(self):
        """Returns radius of the sphere."""
        return self._radius

    @property
    def grid(self):
        """Returns spatial grid option."""
        return self._grid

    @property
    def extend(self):
        """True is spatial grid contains longitudes 0 and 360."""
        return self._extend

    @property
    def normalization(self):
        """Spherical harmonic normalisation option."""
        return self._normalization

    @property
    def csphase(self):
        """Condon Shortley phase option."""
        return self._csphase

    def _coefficient_to_component_mapping(self):
        # returns a sparse matrix that maps from flattended pyshtools
        # coefficients to component vectors.

        row_dim = (self.lmax + 1) ** 2
        col_dim = 2 * (self.lmax + 1) ** 2

        row = 0
        rows = []
        cols = []
        for l in range(self.lmax + 1):
            col = l * (self.lmax + 1)
            for _ in range(l + 1):
                rows.append(row)
                row += 1
                cols.append(col)
                col += 1

        for l in range(self.lmax + 1):
            col = (self.lmax + 1) ** 2 + l * (self.lmax + 1) + 1
            for _ in range(1, l + 1):
                rows.append(row)
                row += 1
                cols.append(col)
                col += 1

        data = [1] * row_dim
        return coo_array(
            (data, (rows, cols)), shape=(row_dim, col_dim), dtype=int
        ).tocsc()

    def spherical_harmonic_index(self, l, m):
        """Return the component index for given spherical harmonic degree and order."""
        if m >= 0:
            return int(l * (l + 1) / 2) + m
        else:
            offset = int((self.lmax + 1) * (self.lmax + 2) / 2)
            return offset + int((l - 1) * l / 2) - m - 1

    def _to_components_from_coeffs(self, coeffs):
        """Return component vector from coefficient array."""
        f = coeffs.flatten(order="C")
        return self._sparse_coeffs_to_component @ f

    def _to_components_from_SHCoeffs(self, ulm):
        """Return component vector from SHCoeffs object."""
        return self._to_components_from_coeffs(ulm.coeffs)

    def _to_components_from_SHGrid(self, u):
        """Return component vector from SHGrid object."""
        ulm = u.expand(normalization=self.normalization, csphase=self.csphase)
        return self._to_components_from_SHCoeffs(ulm)

    def _from_components_to_SHCoeffs(self, c):
        """Return SHCoeffs object from its component vector."""
        f = self._sparse_coeffs_to_component.T @ c
        coeffs = f.reshape((2, self.lmax + 1, self.lmax + 1))
        return sh.SHCoeffs.from_array(
            coeffs, normalization=self.normalization, csphase=self.csphase
        )

    def _from_components_to_SHGrid(self, c):
        """Return SHGrid object from its component vector."""
        ulm = self._from_components_to_SHCoeffs(c)
        return ulm.expand(grid=self.grid, extend=self.extend)

    def _degree_dependent_scaling_to_diagonal_matrix(self, f):
        values = np.zeros(self.dim)
        i = 0
        for l in range(self.lmax + 1):
            j = i + l + 1
            values[i:j] = f(l)
            i = j
        for l in range(1, self.lmax + 1):
            j = i + l
            values[i:j] = f(l)
            i = j
        return diags([values], [0])


class Sobolev(SHToolsHelper, HilbertSpace):
    """Sobolev spaces on a two-sphere as an instance of HilbertSpace."""

    def __init__(
        self, lmax, order, scale, /, *, vector_as_SHGrid=True, radius=1, grid="DH"
    ):
        """
        Args:
            lmax (int): Truncation degree for spherical harmoincs.
            order (float): Order of the Sobolev space.
            scale (float): Non-dimensional length-scale for the space.
            vector_as_SHGrid (bool): If true, elements of the space are
                instances of the SHGrid class, otherwise they are SHCoeffs.
            radius (float): Radius of the two sphere.
            grid (str): pyshtools grid type.
        """
        self._order = order
        self._scale = scale
        SHToolsHelper.__init__(self, lmax, radius=radius, grid=grid)
        self._metric_tensor = self._degree_dependent_scaling_to_diagonal_matrix(
            self._sobolev_function
        )
        self._inverse_metric_tensor = self._degree_dependent_scaling_to_diagonal_matrix(
            lambda l: 1 / self._sobolev_function(l)
        )
        self._vector_as_SHGrid = vector_as_SHGrid
        if vector_as_SHGrid:
            HilbertSpace.__init__(
                self,
                self.dim,
                self._to_components_from_SHGrid,
                self._from_components_to_SHGrid,
                self._inner_product_impl,
                self._to_dual_impl,
                self._from_dual_impl,
            )
        else:
            HilbertSpace.__init__(
                self,
                self.dim,
                self._to_components_from_SHCoeffs,
                self._from_components_to_SHCoeffs,
                self._inner_product_impl,
                self._to_dual_impl,
                self._from_dual_impl,
            )

    # =============================================#
    #                   Properties                 #
    # =============================================#

    @property
    def order(self):
        """Order of the Sobolev space."""
        return self._order

    @property
    def scale(self):
        """Non-dimensional length-scale."""
        return self._scale

    @property
    def laplace_beltrami_operator(self):
        """
        Returns the Laplace Beltrami operator on the space.
        """
        codomain = Sobolev(self.lmax, self.order - 2, self.scale)
        return self.invariant_operator(codomain, lambda l: l * (l + 1) / self.radius**2)

    # ==============================================#
    #                 Public methods                #
    # ==============================================#

    def low_degree_projection(self, truncation_degree, /, *, smoother=None):
        """
        Returns a LinearOperator that maps the space onto a Sobolev space with
        the same parameters but based on a lower truncation degree.
        """
        truncation_degree = (
            truncation_degree if truncation_degree <= self.lmax else self.lmax
        )
        f = smoother if smoother is not None else lambda l: l

        # construct the spare matrix that performs the coordinate projection.
        row_dim = (truncation_degree + 1) ** 2
        col_dim = (self.lmax + 1) ** 2

        row = 0
        col = 0
        rows = []
        cols = []
        data = []
        for l in range(self.lmax + 1):
            fac = f(l)
            for _ in range(l + 1):
                if l <= truncation_degree:
                    rows.append(row)
                    row += 1
                    cols.append(col)
                    data.append(fac)
                col += 1

        for l in range(truncation_degree + 1):
            fac = f(l)
            for _ in range(1, l + 1):
                rows.append(row)
                row += 1
                cols.append(col)
                data.append(fac)
                col += 1

        smat = coo_array(
            (data, (rows, cols)), shape=(row_dim, col_dim), dtype=int
        ).tocsc()

        codomain = Sobolev(
            truncation_degree,
            self.order,
            self.scale,
            vector_as_SHGrid=self._vector_as_SHGrid,
            radius=self.radius,
            grid=self._grid,
        )

        def mapping(u):
            uc = self.to_components(u)
            vc = smat @ uc
            return codomain.from_components(vc)

        def adjoint_mapping(v):
            vc = codomain.to_components(v)
            uc = smat.T @ vc
            return self.from_components(uc)

        return LinearOperator(self, codomain, mapping, adjoint_mapping=adjoint_mapping)

    def dirac(self, latitude, longitude, /, *, degrees=True):
        """
        Returns the Dirac measure at the given point as an
        instance of LinearForm.

        Args:
            latitude (float): Latitude for the Dirac measure.
            longitude (float): Longitude for the Dirac measure.
            degrees (bool): If true, angles in degrees, otherwise
                they are in radians.
        Returns:
            LinearForm: The Dirac measure as a linear form on the space.

        Notes:
            The form is only well-defined mathematically if the order of the
            space is greater than 1.
        """
        if degrees:
            colatitude = 90 - latitude
        else:
            colatitude = np.pi / 2 - latitude
        coeffs = sh.expand.spharm(
            self.lmax,
            colatitude,
            longitude,
            normalization="ortho",
            degrees=degrees,
        )
        c = self._to_components_from_coeffs(coeffs)
        return self.dual.from_components(c)

    def dirac_representation(self, latitude, longitude, /, *, degrees=True):
        """
        Returns representation of the Dirac measure at the given point
        within the Sobolev space.

        Args:
            latitude (float): Latitude for the Dirac measure.
            longitude (float): Longitude for the Dirac measure.
            degrees (bool): If true, angles in degrees, otherwise
                they are in radians.
        Returns:
            LinearForm: The Dirac measure as a linear form on the space.

        Notes:
            The form is only well-defined mathematically if the order of the
            space is greater than 1.
        """
        up = self.dirac(latitude, longitude, degrees=degrees)
        return self.from_dual(up)

    def point_evaluation_operator(self, lats, lons, /, *, degrees=True):
        """
        Returns a LinearOperator instance that maps an element of the
        space to its values at the given set of latitudes and longitudes.

        Args:
            lats (ArrayLike): Array of latitudes.
            lons (ArrayLike): Array of longitudes.
            degrees (bool): If true, angles input as degrees, else in radians.

        Returns:
            LinearOperator: The operator on the space.

        Raises:
            ValueError: If the number of latitudes and longitudes are not equal.

        Notes:
            The operator is only well-defined mathematically if the order of the
            space is greater than 1.
        """

        if lats.size != lons.size:
            raise ValueError("Must have the same number of latitudes and longitudes.")
        codomain = EuclideanSpace(lats.size)

        def mapping(u):
            ulm = u.expand(normalization=self.normalization, csphase=self.csphase)
            return ulm.expand(lat=lats, lon=lons, degrees=degrees)

        def dual_mapping(vp):
            cvp = codomain.dual.to_components(vp)
            cup = np.zeros(self.dim)
            for c, lat, lon in zip(cvp, lats, lons):
                dirac = self.dirac(lat, lon, degrees=degrees)
                cup += c * self.dual.to_components(dirac)
            return LinearForm(self, components=cup)

        return LinearOperator(self, codomain, mapping, dual_mapping=dual_mapping)

    def invariant_operator(self, codomain, f):
        """
        Returns a rotationally invariant linear operator from the
        Sobolev space to another one.

        Args:
            codoaim (Sobolev): The codomain of the operator.
            f (callable): The degree-dependent scaling-function.

        Returns:
            LinearOperator: The linear operator.

        Raises:
            ValueError: If the codomain is not another Sobolev on a two-sphere.
        """
        if not isinstance(codomain, Sobolev):
            raise ValueError("Codomain must be another Sobolev space on a sphere.")
        matrix = self._degree_dependent_scaling_to_diagonal_matrix(f)
        if codomain == self:

            def mapping(x):
                return self.from_components(matrix @ self.to_components(x))

            return LinearOperator.self_adjoint(self, mapping)
        else:

            def mapping(x):
                return codomain.from_components(matrix @ self.to_components(x))

            def dual_mapping(yp):
                return self.dual.from_components(
                    matrix @ codomain.dual.to_components(yp)
                )

            return LinearOperator(self, codomain, mapping, dual_mapping=dual_mapping)

    def invariant_gaussian_measure(self, f, /, *, expectation=None):
        """
        Returns an invariant Gaussian measure on the space.

        Args:
            f (callable): Degree-dependent scaling function that defines the
                covariance operator.
            expectation (Sobolev vector | None): The expected value of the measure.

        Returns:
            GaussianMeasure: The Gaussian measure on the space.

        """

        def g(l):
            return np.sqrt(f(l) / (self.radius**2 * self._sobolev_function(l)))

        def h(l):
            return np.sqrt(self.radius**2 * self._sobolev_function(l) * f(l))

        matrix = self._degree_dependent_scaling_to_diagonal_matrix(g)
        adjoint_matrix = self._degree_dependent_scaling_to_diagonal_matrix(h)
        domain = EuclideanSpace(self.dim)

        def mapping(c):
            return self.from_components(matrix @ c)

        def adjoint_mapping(u):
            return adjoint_matrix @ self.to_components(u)

        inverse_matrix = self._degree_dependent_scaling_to_diagonal_matrix(
            lambda l: 1 / g(l)
        )

        inverse_adjoint_matrix = self._degree_dependent_scaling_to_diagonal_matrix(
            lambda l: 1 / h(l)
        )

        def inverse_mapping(u):
            return inverse_matrix @ self.to_components(u)

        def inverse_adjoint_mapping(c):
            return self.from_components(inverse_adjoint_matrix @ c)

        covariance_factor = LinearOperator(
            domain, self, mapping, adjoint_mapping=adjoint_mapping
        )

        inverse_covariance_factor = LinearOperator(
            self, domain, inverse_mapping, adjoint_mapping=inverse_adjoint_mapping
        )

        return GaussianMeasure(
            covariance_factor=covariance_factor,
            inverse_covariance_factor=inverse_covariance_factor,
            expectation=expectation,
        )

    def sobolev_gaussian_measure(self, order, scale, amplitude, /, *, expectation=None):
        """
        Returns an invariant Gaussian measure on the space whose covariance
        operator takes the Sobolev form:

        Args:
            order (float): The Sobolev order for the covariance.
            scale (float): The non-dimensional length-scale for the covariance.
            amplitude (float): The standard deviation of point values.
            expectation (Sobolev vector | None): The expected value of the measure.

        Returns:
            GaussianMeasures: The Gaussian measure on the space.
        """

        def f(l):
            return (1 + scale**2 * l * (l + 1)) ** (-order)

        return self.invariant_gaussian_measure(
            self._normalise_covariance_function(f, amplitude), expectation=expectation
        )

    def heat_kernel_gaussian_measure(self, scale, amplitude, /, *, expectation=None):
        """
        Returns an invariant Gaussian measure on the space whose covariance
        operator takes the form of a heat kernel.

        Args:
            scale (float): The non-dimensional length-scale for the covariance.
            amplitude (float): The standard deviation of point values.
            expectation (Sobolev vector | None): The expected value of the measure.

        Returns:
            GaussianMeasures: The Gaussian measure on the space.
        """

        def f(l):
            return np.exp(-0.5 * l * (l + 1) * scale**2)

        return self.invariant_gaussian_measure(
            self._normalise_covariance_function(f, amplitude), expectation=expectation
        )

    def sample_variance(self, vectors, /, *, expectation=None):
        """
        Given a list of elements in the space, forms a field
        of point-wise sample variances.

        Args:
            vectors ([vector]): A list of vectors.
            expectation (vector): The expected value that can be provided.
                If not, the sample expectation is used.
        """
        n = len(vectors)
        if expectation is None:
            assert n > 1
            fac = 1 / (n - 1)
            ubar = self.sample_expectation(vectors)
        else:
            assert n > 0
            fac = 1 / n
            ubar = expectation
        uvar = self.zero
        for u in vectors:
            u = (u - ubar) * (u - ubar)
            uvar = self.axpy(fac, u, uvar)
        return uvar

    def sample_std(self, vectors, /, *, expectation=None):
        """
        Given a list of elements in the space, forms a field
        of point-wise sample standard deviations.

        Args:
            vectors ([vector]): A list of vectors.
            expectation (vector): The expected value that can be provided.
                If not, the sample expectation is used.
        """
        ustd = self.sample_variance(vectors, expectation=expectation)
        ustd.data = np.sqrt(ustd.data)
        return ustd

    def plot(self, u, *args, **kwargs):
        """
        Make a simple plot of an element of the space.
        """
        if self._vector_as_SHGrid:
            plt.pcolormesh(u.lons(), u.lats(), u.data, *args, **kwargs)
        else:
            self.plot(u.expand(grid=self.grid, extend=self.extend), *args, **kwargs)

    # ==============================================#
    #                Private methods               #
    # ==============================================#

    def _sobolev_function(self, l):
        # Degree-dependent scaling that defines the Sobolev inner product.
        return (1 + self.scale**2 * l * (l + 1)) ** self.order

    def _inner_product_impl(self, u, v):
        # Implementation of the inner product.
        return self.radius**2 * np.dot(
            self._metric_tensor @ self.to_components(u), self.to_components(v)
        )

    def _to_dual_impl(self, u):
        # Implementation of the mapping to the dual space.
        c = self._metric_tensor @ self.to_components(u) * self.radius**2
        return self.dual.from_components(c)

    def _from_dual_impl(self, up):
        # Implementation of the mapping from the dual space.
        c = self._inverse_metric_tensor @ self.dual.to_components(up) / self.radius**2
        return self.from_components(c)

    def _normalise_covariance_function(self, f, amplitude):
        # Normalise a degree-dependent scaling function, f, so that
        # the associated invariant Gaussian measure has standard deviation
        # for point values equal to amplitude.
        norm = 0
        for l in range(self.lmax + 1):
            norm += (
                f(l)
                * (2 * l + 1)
                / (4 * np.pi * self.radius**2 * self._sobolev_function(l))
            )
        return lambda l: amplitude**2 * f(l) / norm


class Lebesgue(Sobolev):
    """
    L2 on the two-sphere as an instance of HilbertSpace.

    Implemented as a special case of the Sobolev class with order = 0.
    """

    def __init__(self, lmax, /, *, vector_as_SHGrid=True, radius=1, grid="DH"):
        """
        Args:
            lmax (int): Truncation degree for spherical harmoincs.
            vector_as_SHGrid (bool): If true, elements of the space are
                instances of the SHGrid class, otherwise they are SHCoeffs.
            radius (float): Radius of the two sphere.
            grid (str): pyshtools grid type.
        """
        super().__init__(
            lmax,
            0,
            0,
            vector_as_SHGrid=vector_as_SHGrid,
            radius=radius,
            grid=grid,
        )


###############################################################
#                      Utility classes                        #
###############################################################


class LowPassFilter:
    """
    Class implementing a simple Hann-type low-pass filter in
    the spherical harmonic domain
    """

    def __init__(self, lower_degree, upper_degree):
        """
        Args:
            lower_degree (int): Below this degree, the filter returns one.
            upper_degree (int): Above this degree, the filter returns zero.
                Its value between the lower and upper degrees decreases smoothly.
        """
        self._lower_degree = lower_degree
        self._upper_degree = upper_degree

    def __call__(self, l):
        if l <= self._lower_degree:
            return 1
        elif self._lower_degree <= l <= self._upper_degree:
            return 0.5 * (
                1
                - np.cos(
                    np.pi
                    * (self._upper_degree - l)
                    / (self._upper_degree - self._lower_degree)
                )
            )
        else:
            return 0
